@@ -13,6 +13,7 @@ __all__ = [
     'queue_fetch_task',
     'crl_job_results_as_completed',
     'ocsp_job_get_earliest',
+    'complete_certificate_fetch_jobs', 'gather_aia_issuer_urls',
     'ACCEPTABLE_STRICT_CERT_CONTENT_TYPES',
     'ACCEPTABLE_CERT_PEM_ALIASES'
 ]
@@ -23,24 +24,25 @@ logger = logging.getLogger(__name__)
 
 ACCEPTABLE_STRICT_CERT_CONTENT_TYPES = (
     'application/pkix-cert', 'application/pkcs7-mime',
+    'application/x-x509-ca-cert'
 )
 
 ACCEPTABLE_CERT_PEM_ALIASES = (
-    'application/x-pem-file', 'application/x-x509-ca-cert', 'text/plain',
+    'application/x-pem-file', 'text/plain',
 )
 
 
 def unpack_cert_content(response_data: bytes, content_type: str,
                         url: str, permit_pem: bool):
 
-    if content_type == 'application/pkix-cert':
+    if content_type in ('application/pkix-cert', 'application/x-x509-ca-cert'):
         yield x509.Certificate.load(response_data)
     elif permit_pem and (content_type in ACCEPTABLE_CERT_PEM_ALIASES):
         # technically, PEM is not allowed here, but of course some people don't
         # bother following the rules
         if pem.detect(response_data):
             for _, _, data in pem.unarmor(response_data, multiple=True):
-                yield x509.Certificate.load(response_data)
+                yield x509.Certificate.load(data)
         else:
             raise ValueError(
                 "Expected PEM data when extracting certs from "
@@ -96,7 +98,12 @@ def format_ocsp_request(cert: x509.Certificate, issuer: x509.Certificate,
 
 def process_ocsp_response_data(response_data: bytes, *,
                                ocsp_request: ocsp.OCSPRequest, ocsp_url: str):
-    ocsp_response = ocsp.OCSPResponse.load(response_data)
+    try:
+        ocsp_response = ocsp.OCSPResponse.load(response_data)
+    except ValueError:
+        raise errors.OCSPFetchError(
+            'Failed to parse response from OCSP server'
+        )
     status = ocsp_response['response_status'].native
     if status != 'successful':
         raise errors.OCSPValidationError(
@@ -217,3 +224,31 @@ async def ocsp_job_get_earliest(jobs):
         await cancel_all(queue)
         return ocsp_resp
     raise last_e or errors.OCSPFetchError("No OCSP results")
+
+
+def gather_aia_issuer_urls(cert: x509.Certificate):
+    aia_value = cert.authority_information_access_value
+    if aia_value is None:
+        return
+    for entry in aia_value:
+        if entry['access_method'].native == 'ca_issuers':
+            location = entry['access_location']
+            if location.name != 'uniform_resource_identifier':
+                continue
+            url = location.native
+            if url.startswith('http'):
+                yield url
+
+
+async def complete_certificate_fetch_jobs(fetch_jobs):
+    for fetch_job in asyncio.as_completed(fetch_jobs):
+        try:
+            certs_fetched = await fetch_job
+        except errors.CertificateFetchError as e:
+            logger.warning(
+                f'Error during certificate fetch job, skipping... '
+                f'(Error: {e})',
+            )
+            continue
+        for cert in certs_fetched:
+            yield cert
